package org.netkernel.lang.dpml.ast.impl;

import org.netkernel.lang.dpml.ast.*;
import org.netkernel.lang.dpml.endpoint.DPMLUtils;
import org.netkernel.lang.dpml.endpoint.DPMLValueSpace;
import org.netkernel.lang.dpml.endpoint.RuntimeState;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.nkf.impl.NKFResponseReadOnlyImpl;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.RequestScopeClassLoader;
import org.netkernel.request.IResponse;
import org.netkernel.request.impl.ExpiryFactory;
import org.netkernel.request.impl.MetaImpl;

import java.util.Collections;
import java.util.List;


/** A sub request with arguments
 * @author Tony Butterfield
 *
 */
public class SubRequest extends VariableImpl implements IValueVariable
{
	private final String mRequestId;
	private final int mVerb;
	private final List mArguments;
	private final List mHeaders;
	private final String mRepName;
	private final boolean mVarargs;
	
	public SubRequest(String aName, IScope aScope, ILocation aLocation, String aRequestId, int aVerb, List aArguments, List aHeaders, String aRepName, boolean aVarargs)
	{	super(aName,aScope,aLocation);
		mRequestId=aRequestId;
		mVerb=aVerb;
		mArguments=aArguments.size()>0?aArguments:Collections.EMPTY_LIST;
		mHeaders=aHeaders.size()>0?aHeaders:Collections.EMPTY_LIST;
		mRepName=aRepName;
		mVarargs=aVarargs;
	}
	
	public IValue getValue(INKFRequestContext aContext, RuntimeState aState, Class aRepresentationClass) throws NKFException
	{
		if (getScope() instanceof Closure && !this.isAnonymous())
		{
			IVariable cached=aState.getAssignment(getScope(), getName());
			if (cached!=null && cached instanceof IValueVariable)
			{	IValueVariable vv = (IValueVariable)cached;
				IValue v=vv.getValue(aContext, aState, aRepresentationClass);
				Object rep=v.getRepresentation();
				
				if (rep!=null && !aRepresentationClass.isAssignableFrom(rep.getClass()))
				{	//in cache but not right representation
					try
					{	INKFResponse respIn=aContext.createResponseFrom(rep);
						respIn.setHeaders(v.getMeta().getUserMetaData());
						INKFResponseReadOnly transreptResp=aContext.transreptForResponse(respIn, aRepresentationClass);
						v=new ValueImpl(transreptResp.getRepresentation(), ((NKFResponseReadOnlyImpl)transreptResp).getMeta());
						return v;
					}
					catch (NKFException e)
					{	//couldn't transrept so let's just do the work again
					}
				}
				else
				{	return v;
				}
			}
		}
		
		INKFRequest req = aContext.createRequest(mRequestId);
		req.setVerb(mVerb);
		if (mRepName!=null)
		{	RequestScopeClassLoader cl = new RequestScopeClassLoader(aContext.getKernelContext().getRequestScope());
			try
			{	Class repClass = cl.loadClass(mRepName);
				req.setRepresentationClass(repClass);
			}
			catch (Exception e)
			{	throw aContext.createFormattedException("EX_DPML_RUNTIME", "MSG_REP_NOT_FOUND", DPMLUtils.formatLocationMessage(this,aContext), e, mRepName);
			}
		}
		else
		{	req.setRepresentationClass(aRepresentationClass);
		}
		
		DPMLValueSpace vs=new DPMLValueSpace(mArguments.size(),aContext.getKernelContext().getRequestScope());
		for (int i=0; i<mArguments.size(); i++)
		{	IArgument arg=(IArgument)mArguments.get(i);
			String name=arg.getName();
			
			if (name.equals(IArgument.ARG_PRIMARY))
			{
				IValueVariable var=(IValueVariable) DPMLUtils.processArgument(arg,aContext,aState,true);
				IValue value=var.getValue(aContext,aState,Object.class);
				req.addPrimaryArgumentFromResponse(DPMLUtils.createResponseWithMetaData(value,aContext));
			}
			else
			{	IVariable var=arg.getVariable();
				var= DPMLUtils.processArgument(arg,aContext,aState,false);
				
				if (var instanceof IIdentifierVariable)
				{	String identifier = ((IIdentifierVariable)var).getIdentifier(aContext,aState);
					if (((IIdentifierVariable)var).isAbsolute())
					{	req.addArgumentAsAbsolute( name,  identifier);
					}
					else
					{	req.addArgument( name,  identifier);
					}
				}
				else if (var instanceof IValueVariable)
				{	if (var instanceof LiteralValue)
					{	LiteralValue lv=(LiteralValue)var;
						IValue v=lv.getValue(aContext, aState, aRepresentationClass);
						String identifier=vs.createPBVIdentifier(name);
						req.addArgumentAsAbsolute(name, identifier);
						vs.put(identifier, v.getRepresentation(), v.getMeta());
					}
					else
					{	// subrequest/augmented closure/closure/sequence
						String identifier=vs.createPBVIdentifier(name);
						req.addArgumentAsAbsolute(name, identifier);
						vs.addLazyEvaluation(identifier,(IValueVariable)var,aState);
					}
				}
			}
		}
		
		//varargs
		if (mVarargs)
		{
			RuntimeState s=aState;
			while (s!=null && s.getArguments()==null)
			{	s=s.getParent();
			}
			PairList args=s.getArguments();
			if (args!=null && args.size()>0)
			{
				for (int i=0; i<args.size(); i++)
				{	String name=(String)args.getValue1(i);
					boolean alreadyUsed=name.equals("verb") || name.equals("arguments");
					if (!alreadyUsed)
					{	for (int j=0; j<mArguments.size(); j++)
						{	IArgument arg=(IArgument)mArguments.get(j);
							if (arg.getName().equals(name))
							{	alreadyUsed=true;
								break;
							}
						}
					}
					if (!alreadyUsed)
					{	ArgumentReference argRef=new ArgumentReference(null, getScope(), getLocation(), name, INKFRequestReadOnly.VERB_SOURCE);
						IVariable var=DPMLUtils.dereference(argRef, s, aContext);						
						if (var instanceof IIdentifierVariable)
						{	String identifier = ((IIdentifierVariable)var).getIdentifier(aContext,aState);
							if (((IIdentifierVariable)var).isAbsolute())
							{	req.addArgumentAsAbsolute( name,  identifier);
							}
							else
							{	req.addArgument( name,  identifier);
							}
						}
						else if (var instanceof IValueVariable)
						{	if (var instanceof LiteralValue)
							{	LiteralValue lv=(LiteralValue)var;
								IValue v=lv.getValue(aContext, aState, aRepresentationClass);
								String identifier=vs.createPBVIdentifier(name);
								req.addArgumentAsAbsolute(name, identifier);
								vs.put(identifier, v.getRepresentation(), v.getMeta());
							}
							else
							{	// subrequest/augmented closure/closure/sequence
								String identifier=vs.createPBVIdentifier(name);
								req.addArgumentAsAbsolute(name, identifier);
								vs.addLazyEvaluation(identifier,(IValueVariable)var,aState);
							}
						}
					}
				}
			}
		}
		
		
		if (vs.getSize()!=0)
		{	try
			{	vs.onCommissionSpace(aContext.getKernelContext().getKernel());
			} catch (Exception e) {/* can't happen for a valuespace */}
			req.injectDurableRequestScope(vs);
		}
		
		for (int i=0; i<mHeaders.size(); i++)
		{	IRequestHeader header=(IRequestHeader)mHeaders.get(i);
			String name=header.getName();
			Object rep= DPMLUtils.processHeader(header,aContext,aState);
			req.setHeader(name, rep, header.isSticky());
		}
		
		INKFResponseReadOnly resp=aContext.issueRequestForResponse(req);
		if (!this.isAnonymous())
		{	LiteralValue toCache=new LiteralValue(getName(), getScope(), getLocation(), resp.getRepresentation(),((NKFResponseReadOnlyImpl)resp).getKernelResponse().getMeta());
			aState.putAssignment(getScope(), getName(), toCache);
		}
		return new ValueImpl(resp.getRepresentation(),((NKFResponseReadOnlyImpl)resp).getKernelResponse().getMeta());
	}
	
	public String toString()
	{	return "SubRequest";
	}	
}
